/* rclowpass.c
 * ------------------------------------------------------------------------
 *
 * Copyright (C) 2000 Stefan Fendt <stefan@lionfish.ping.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 * ------------------------------------------------------------------------
 *
 * The following function provides us a 2nd order lowpass-filter with 
 * resonance. This is a simulation of an active RC-lowpass. That is the 
 * ammount of resonance is controling the filter curve, too. With too less 
 * resonance you will not achieve a 24dB filter but only 12dB. As this is
 * the normal behavior of analog filters this is intentional.
 *
 * cutoff-range is 0...1
 * resonace-range is 0...~300
 *
 * I have tested it with resonance-values up to 300 and everything worked
 * fine, but I can't say if it will be with even higher values although I 
 * don't expect problems with that.
 * As I wrote this simulation for a synthesizer (where I don't need absolute
 * frequencies) cutoff-values are not given as frequencies, even if the 
 * mathematical would be done quite easiely... If you need that, look into a 
 * good book on analouge filters and be warned that it is nonlinear.
 *
 * NOTE 28th of JAN: I need to speed up the things sometimes ...
 * NOTE 05th of FEB: Changed resonance-model...
 *                   changed filter, too ... now ~36dB
 * 
 */

#include <stdio.h>
#include <math.h>
#define LEN 512

float rc_lowpass(float cutoff, float resonance, float invalue)
{
  static float lp1,lp2,lp3,hp1,feedback,l_in;
  float value,value0,value1,value2,value3,last;

  resonance*=(cutoff*0.3+1.3);

  value0=last;
  value1=last+0.25*(invalue-last);
  value2=last+0.50*(invalue-last);
  value3=last+0.75*(invalue-last);

  if(feedback> 32000) feedback=32000;
  if(feedback<-32000) feedback=-32000;

  value = value0+feedback*resonance;
  
  lp1  = value*cutoff   + lp1*(1-cutoff) ;
  lp2  = lp1*cutoff     + lp2*(1-cutoff) ;
  lp3  = lp2*cutoff     + lp3*(1-cutoff) ;

  hp1=lp3-lp1;

  feedback=-hp1;

  if(feedback> 32000) feedback=32000;
  if(feedback<-32000) feedback=-32000;

  value = value1+feedback*resonance;
  
  lp1  = value*cutoff   + lp1*(1-cutoff) ;
  lp2  = lp1*cutoff     + lp2*(1-cutoff) ;
  lp3  = lp2*cutoff     + lp3*(1-cutoff) ;

  hp1=lp3-lp1;

  feedback=-hp1;

  if(feedback> 32000) feedback=32000;
  if(feedback<-32000) feedback=-32000;

  value = value2+feedback*resonance;
  
  lp1  = value*cutoff   + lp1*(1-cutoff) ;
  lp2  = lp1*cutoff     + lp2*(1-cutoff) ;
  lp3  = lp2*cutoff     + lp3*(1-cutoff) ;

  hp1=lp3-lp1;

  feedback=-hp1;

  if(feedback> 32000) feedback=32000;
  if(feedback<-32000) feedback=-32000;

  value = value3+feedback*resonance;
  
  lp1  = value*cutoff   + lp1*(1-cutoff) ;
  lp2  = lp1*cutoff     + lp2*(1-cutoff) ;
  lp3  = lp2*cutoff     + lp3*(1-cutoff) ;

  hp1=lp3-lp1;

  feedback=-hp1;

  last=invalue;

  return lp3;             
}


float rc_highpass(float cutoff, float resonance, float invalue)
{
  static float lp1,lp2,lp3,lp4,hp1,hp2,feedback,feedback2,last;
  float value,value2;
  float invalue0,invalue1,invalue2,invalue3;


  invalue0=last;
  invalue1=last+0.25*(invalue-last);
  invalue2=last+0.50*(invalue-last);
  invalue3=last+0.75*(invalue-last);

  value = invalue0-feedback*resonance;
  
  lp1  = value*cutoff   + lp1*(1-cutoff) ;
  lp2  = lp1*cutoff     + lp2*(1-cutoff) ;

  hp1=lp2-value;

  feedback=hp1*(cutoff)+feedback*(1-cutoff);


  value2 = hp1*2-feedback2*resonance*2.5;
  
  lp3  = value2*cutoff   + lp3*(1-cutoff) ;
  lp4  = lp1*cutoff     + lp4*(1-cutoff) ;

  hp2=lp3-value2;

  feedback2=hp2*(cutoff)+feedback2*(1-cutoff);


  value = invalue1-feedback*resonance;
  
  lp1  = value*cutoff   + lp1*(1-cutoff) ;
  lp2  = lp1*cutoff     + lp2*(1-cutoff) ;

  hp1=lp2-value;

  feedback=hp1*(cutoff)+feedback*(1-cutoff);


  value2 = hp1*2-feedback2*resonance*2.5;
  
  lp3  = value2*cutoff   + lp3*(1-cutoff) ;
  lp4  = lp1*cutoff     + lp4*(1-cutoff) ;

  hp2=lp3-value2;

  feedback2=hp2*(cutoff)+feedback2*(1-cutoff);


  value = invalue2-feedback*resonance;
  
  lp1  = value*cutoff   + lp1*(1-cutoff) ;
  lp2  = lp1*cutoff     + lp2*(1-cutoff) ;

  hp1=lp2-value;

  feedback=hp1*(cutoff)+feedback*(1-cutoff);


  value2 = hp1*2-feedback2*resonance*2.5;
  
  lp3  = value2*cutoff   + lp3*(1-cutoff) ;
  lp4  = lp1*cutoff     + lp4*(1-cutoff) ;

  hp2=lp3-value2;

  feedback2=hp2*(cutoff)+feedback2*(1-cutoff);


  value = invalue3-feedback*resonance;
  
  lp1  = value*cutoff   + lp1*(1-cutoff) ;
  lp2  = lp1*cutoff     + lp2*(1-cutoff) ;

  hp1=lp2-value;

  feedback=hp1*(cutoff)+feedback*(1-cutoff);


  value2 = hp1*2-feedback2*resonance*2.5;
  
  lp3  = value2*cutoff   + lp3*(1-cutoff) ;
  lp4  = lp1*cutoff     + lp4*(1-cutoff) ;

  hp2=lp3-value2;

  feedback2=hp2*(cutoff)+feedback2*(1-cutoff);

  last=invalue;

  return hp2;             
}

float lp(int n, float in,float ff)
{
    #define PI 3.1415926
    #define m 32000

    static float l0[8],h0[8],b0[8];
    static float l1[8],h1[8],b1[8];
    float q;
    float f;

    /* feedback factor "q" */

    q=1.125;

    /* frequency calculation */

    f= 2*sin (PI * ff / 44100);
    f= (f>1)? 1:f;
    f= (f<0)? 0:f;

    /* stage 1 */

    in=in+b0[n]*q;
    l0[n]=l0[n]*(1-f)+in*(f);
    h0[n]=in-l0[n];
    b0[n]=b0[n]*(1-f)+h0[n]*(f);

    /* stage 2 */

    in=l0[n]+b1[n]*q;
    l1[n]=l1[n]*(1-f)+in*(f);
    h1[n]=in-l1[n];
    b1[n]=b1[n]*(1-f)+h1[n]*(f);

    /* limit filter registers */

    l0[n] = (l0[n]>+m)? +m:l0[n];
    l0[n] = (l0[n]<-m)? -m:l0[n];
    l1[n] = (l1[n]>+m)? +m:l1[n];
    l1[n] = (l1[n]<-m)? -m:l1[n];

    h0[n] = (h0[n]>+m)? +m:h0[n];
    h0[n] = (h0[n]<-m)? -m:h0[n];
    h1[n] = (h1[n]>+m)? +m:h1[n];
    h1[n] = (h1[n]<-m)? -m:h1[n];

    b0[n] = (b0[n]>+m)? +m:b0[n];
    b0[n] = (b0[n]<-m)? -m:b0[n];
    b1[n] = (b1[n]>+m)? +m:b1[n];
    b1[n] = (b1[n]<-m)? -m:b1[n];

    return l1[n];
}

float bp(int n, float in,float ff)
{
    #define PI 3.1415926
    #define m 32000

    static float l0[8],h0[8],b0[8];
    static float l1[8],h1[8],b1[8];
    float q;
    float f;

    /* feedback factor "q" */

    q=1.8;

    /* frequency calculation */

    f= 2*sin (PI * ff / 44100);
    f= (f>1)? 1:f;
    f= (f<0)? 0:f;

    /* stage 1 */

    in=in+b0[n]*q;
    l0[n]=l0[n]*(1-f)+in*(f);
    h0[n]=in-l0[n];
    b0[n]=b0[n]*(1-f)+h0[n]*(f);

    /* stage 2 */

    in=b0[n]+b1[n]*q;
    l1[n]=l1[n]*(1-f)+in*(f);
    h1[n]=in-l1[n];
    b1[n]=b1[n]*(1-f)+h1[n]*(f);

    /* limit filter registers */

    l0[n] = (l0[n]>+m)? +m:l0[n];
    l0[n] = (l0[n]<-m)? -m:l0[n];
    l1[n] = (l1[n]>+m)? +m:l1[n];
    l1[n] = (l1[n]<-m)? -m:l1[n];

    h0[n] = (h0[n]>+m)? +m:h0[n];
    h0[n] = (h0[n]<-m)? -m:h0[n];
    h1[n] = (h1[n]>+m)? +m:h1[n];
    h1[n] = (h1[n]<-m)? -m:h1[n];

    b0[n] = (b0[n]>+m)? +m:b0[n];
    b0[n] = (b0[n]<-m)? -m:b0[n];
    b1[n] = (b1[n]>+m)? +m:b1[n];
    b1[n] = (b1[n]<-m)? -m:b1[n];

    return b1[n];
}

float vocal_formant (float input, char vocals[], float time_ms )
{
    #define _a_ 0
    #define _e_ 1
    #define _i_ 2
    #define _o_ 3
    #define _u_ 4

    float t;
    char vowel;
    float formants[5][3]=
    {
	{730,1090,2440},
	{530,1840,2480},
	{270,2290,3010},
	{570, 840,2410},
	{300, 870,2240}
    };

    static float f1=730 ,f1_=730;
    static float f2=1090,f2_=1090;
    static float f3=2440,f3_=2440;

    static int c=-2;

    static int n=0;
    input/=100;

    c++;

    t=2*sin(PI*(500/time_ms)/44100);
    f1=f1*(1-t)+f1_*(t);
    f2=f2*(1-t)+f2_*(t);
    f3=f3*(1-t)+f3_*(t);

    if(c<0 || c>=((float)44100/(1000/time_ms)))
    {
	static int str_pos=0;
	c=0;
	
	vowel=vocals[str_pos++];
	if(vowel==0)
	{
	    str_pos=0;
	    vowel=vocals[str_pos++];
	}

	if(vowel=='a') vowel=_a_;
	if(vowel=='e') vowel=_e_;
	if(vowel=='i') vowel=_i_;
	if(vowel=='o') vowel=_o_;
	if(vowel=='u') vowel=_u_;

	f1_=formants[(int)vowel][0];
	f2_=formants[(int)vowel][1];
	f3_=formants[(int)vowel][2];
    }

/* beware of the following null-constructs. DO NOT TOUCH. 
 * They are important !!!
 */

    input =
	/* lowpass */
	/* let base-frequencies shine through... :) */

	lp(0,input,f1/16)*0+
	lp(0,input,f1/16)*0+
	lp(0,input,f1/16)*0+
	lp(0,input,f1/16)+

	/* first formant */
	bp(0,input,f1/4)*0+
	bp(0,input,f1/4)*0+
	bp(0,input,f1/4)*0+
	bp(0,input,f1/4)+

	/* second formant */
	bp(1,input,f2/4)*0+
	bp(1,input,f2/4)*0+
	bp(1,input,f2/4)*0+
	bp(1,input,f2/4)+

	/* third formant */
	bp(2,input,f3/4)*0+
	bp(2,input,f3/4)*0+
	bp(2,input,f3/4)*0+
	bp(2,input,f3/4);

    return input*10;
}









